﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using MathNet.Numerics.Interpolation;

namespace HeuristicLab.Algorithms.DataAnalysis.Experimental {
  [Item("Splines", "")]
  [Creatable(CreatableAttribute.Categories.DataAnalysisRegression, Priority = 102)]
  [StorableClass]
  public sealed class Splines : FixedDataAnalysisAlgorithm<IRegressionProblem> {
    [StorableConstructor]
    private Splines(bool deserializing) : base(deserializing) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
    }

    private Splines(Splines original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new Splines(this, cloner);
    }

    public Splines()
      : base() {
      var validTypes = new ItemSet<StringValue>(
        new[] {
        "Monotone", "Akima", "Catmull-Rom", "Cubic", "Linear",
          "Cubic - Natural (Math.NET)",
          "Polynomial (Math.NET)",
          "Rational (Math.NET)",
          "LogLinear (Math.NET)",
          "Common (Math.NET)",
      }.Select(s => new StringValue(s)));

      Parameters.Add(new ConstrainedValueParameter<StringValue>("Type", "The type of spline (as supported by alglib)", validTypes, validTypes.First()));
      Problem = new RegressionProblem();
    }


    protected override void Run(CancellationToken cancellationToken) {
      double[,] inputMatrix = Problem.ProblemData.Dataset.ToArray(Problem.ProblemData.AllowedInputVariables.Concat(new string[] { Problem.ProblemData.TargetVariable }), Problem.ProblemData.TrainingIndices);
      if (inputMatrix.Cast<double>().Any(x => double.IsNaN(x) || double.IsInfinity(x)))
        throw new NotSupportedException("Splines does not support NaN or infinity values in the input dataset.");

      var inputVars = Problem.ProblemData.AllowedInputVariables.ToArray();
      if (inputVars.Length > 3) throw new ArgumentException();

      var y = Problem.ProblemData.TargetVariableTrainingValues.ToArray();
      if (inputVars.Length == 1) {

        var x = Problem.ProblemData.Dataset.GetDoubleValues(inputVars.First(), Problem.ProblemData.TrainingIndices).ToArray();
        alglib.spline1dinterpolant c;
        var type = ((StringValue)Parameters["Type"].ActualValue).Value;
        switch (type) {
          case "Monotone":
            alglib.spline1dbuildmonotone(x, y, out c); 
            AddAlglibSplineResult(c, inputVars);
            break;
          case "Akima":
            alglib.spline1dbuildakima(x, y, out c); AddAlglibSplineResult(c, inputVars);
            break;
            ;
          case "Catmull-Rom":
            alglib.spline1dbuildcatmullrom(x, y, out c); AddAlglibSplineResult(c, inputVars);
            break;

          case "Cubic":
            alglib.spline1dbuildcubic(x, y, out c); AddAlglibSplineResult(c, inputVars);
            break;

          case "Linear":
            alglib.spline1dbuildlinear(x, y, out c); AddAlglibSplineResult(c, inputVars);
            break;
          case "Cubic - Natural (Math.NET)": {
              var spline = MathNet.Numerics.Interpolation.CubicSpline.InterpolateNatural(x, y);
              AddMathNetSplineResult(spline, inputVars);
              break;
            }
          case "Common (Math.NET)": {
              var spline = MathNet.Numerics.Interpolate.Common(x, y);
              AddMathNetSplineResult(spline, inputVars);
              break;
            }
          case "LogLinear (Math.NET)": {
              var spline = MathNet.Numerics.Interpolate.LogLinear(x, y);
              AddMathNetSplineResult(spline, inputVars);
              break;
            }
          case "Polynomial (Math.NET)": {
              var spline = MathNet.Numerics.Interpolate.Polynomial(x, y);
              AddMathNetSplineResult(spline, inputVars);
              break;
            }
          case "Rational (Math.NET)": {
              var spline = MathNet.Numerics.Interpolate.RationalWithoutPoles(x, y);
              AddMathNetSplineResult(spline, inputVars);
              break;
            }
          default: throw new NotSupportedException();
        }

      }
    }

    private void AddAlglibSplineResult(alglib.spline1dinterpolant c, string[] inputVars) {
      Results.Add(new Result("Solution", new RegressionSolution(new Spline1dModel(c, Problem.ProblemData.TargetVariable, inputVars),
        (IRegressionProblemData)Problem.ProblemData.Clone())));

    }
    private void AddMathNetSplineResult(IInterpolation c, string[] inputVars) {
      Results.Add(new Result("Solution", new RegressionSolution(new MathNetSplineModel(c, Problem.ProblemData.TargetVariable, inputVars),
        (IRegressionProblemData)Problem.ProblemData.Clone())));

    }
  }


  // UNFINISHED
  public class Spline1dModel : NamedItem, IRegressionModel {
    private alglib.spline1dinterpolant interpolant;

    public string TargetVariable { get; set; }

    public IEnumerable<string> VariablesUsedForPrediction { get; private set; }

    public event EventHandler TargetVariableChanged;

    public Spline1dModel(Spline1dModel orig, Cloner cloner) : base(orig, cloner) {
      this.TargetVariable = orig.TargetVariable;
      this.VariablesUsedForPrediction = orig.VariablesUsedForPrediction.ToArray();
      this.interpolant = (alglib.spline1dinterpolant)orig.interpolant.make_copy();
    }
    public Spline1dModel(alglib.spline1dinterpolant interpolant, string targetVar, string[] inputs) : base("SplineModel", "SplineModel") {
      this.interpolant = interpolant;
      this.TargetVariable = targetVar;
      this.VariablesUsedForPrediction = inputs;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Spline1dModel(this, cloner);
    }

    public IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, (IRegressionProblemData)problemData.Clone());
    }

    public IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      foreach (var x in dataset.GetDoubleValues(VariablesUsedForPrediction.First(), rows)) {
        yield return alglib.spline1dcalc(interpolant, x);
      }
    }
  }


  // UNFINISHED
  public class MathNetSplineModel : NamedItem, IRegressionModel {
    private IInterpolation interpolant;

    public string TargetVariable { get; set; }

    public IEnumerable<string> VariablesUsedForPrediction { get; private set; }

    public event EventHandler TargetVariableChanged;

    public MathNetSplineModel(MathNetSplineModel orig, Cloner cloner) : base(orig, cloner) {
      this.TargetVariable = orig.TargetVariable;
      this.VariablesUsedForPrediction = orig.VariablesUsedForPrediction.ToArray();
      this.interpolant = orig.interpolant; // TODO COPY!
    }
    public MathNetSplineModel(IInterpolation interpolant, string targetVar, string[] inputs) : base("SplineModel", "SplineModel") {
      this.interpolant = interpolant;
      this.TargetVariable = targetVar;
      this.VariablesUsedForPrediction = inputs;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MathNetSplineModel(this, cloner);
    }

    public IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, (IRegressionProblemData)problemData.Clone());
    }

    public IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      foreach (var x in dataset.GetDoubleValues(VariablesUsedForPrediction.First(), rows)) {
        yield return interpolant.Interpolate(x);
      }
    }
  }
}
